#!/usr/bin/env python

import psyco
psyco.full()

import scipy
from scipy import *
import string
import time
from tools import *
from copy import deepcopy

def distance_sum(data, kernels):
  sums = 0
  for point in data:    
    min_diff = None
    for kernel in kernels:
      diff = distance(point, kernel)
      if (min_diff == None or diff < min_diff):
        min_diff = diff
    sums += min_diff
  return sums

def distance_mean(v):
  if (len(v) < 2):
    return 0.
  total_dist = 0.
  nb_dist = 0
  for i in range(0, len(v) - 1):
    for j in range(i + 1, len(v)):
      nb_dist += 1
      total_dist += sqrt(distance(v[i], v[j]))
  total_dist /= nb_dist
  return total_dist

def distance_min(v):
  min_dist = None
  for i in range(0, len(v) - 1):
    for j in range(i + 1, len(v)):
      if (min_dist == None or distance(v[i], v[j]) < min_dist):
        min_dist = distance(v[i], v[j])
  return min_dist

def em(data, alphas, means, variations):
  """Expectation Maximization algorithm"""

  # Store p(c = k | xi) for each xi
  p_class_samples = zeros((len(data), len(alphas)), 'd')
  p_class_norm = 0.
  for i in range(0, len(data)):
    p_class_norm = 0
    for k in range(0, len(alphas)):
      v = alphas[k] * gaussian(means[k], variations[k], array(data[i]))
      p_class_samples[i][k] = v
      p_class_norm += v
    p_class_samples[i] = [x / p_class_norm for x in p_class_samples[i]]

  # Compute the new values of alphas
  for k in range(0, len(alphas)):
    alphas[k] = 0.
    for i in range(0, len(data)):
      alphas[k] += p_class_samples[i][k]
    alphas[k] /= len(data)

  # Store normalization terms for means and variations
  p_class_samples_sum = zeros((len(alphas)), 'd')
  for k in range(0, len(alphas)):
    for i in range(0, len(data)):
      p_class_samples_sum[k] += p_class_samples[i][k]
  
  # Update means
  for k in range(0, len(means)):
    means[k] = 0
    for i in range(0, len(data)):
      means[k] += data[i] * p_class_samples[i][k]
    means[k] /= p_class_samples_sum[k]

  # Update variations
  for k in range(0, len(variations)):
    oldvar = variations[k]
    variations[k] = mat(zeros((len(means[0]), len(means[0])), 'd'))
    for i in range(0, len(data)):
      variations[k] += p_class_samples[i][k] \
                       * mat(data[i] - means[k]).T \
                       * mat(data[i] - means[k])
    variations[k] /= p_class_samples_sum[k]
    # Variations is not inversible, one dimension is flat.
    # Keep the old matrix.
    if (linalg.det(variations[k]) < 0.001):
      variations[k] = oldvar

def kmeans(data, kernels, nb_it):
  for it in range(0, nb_it):
    points_sum = zeros((len(kernels), len(kernels[0])), 'd')
    points_number = zeros(len(kernels), 'd')
    for x in data:
      min_dist = None
      min_kernel = None
      for k in range(0, len(kernels)):
        dist = distance(x, kernels[k])
        if (min_dist == None or dist < min_dist):
          min_dist = dist
          min_kernel = k
      points_sum[min_kernel] += x
      points_number[min_kernel] += 1
    for i in range(0, len(kernels)):
      kernels[i] = points_sum[i] / points_number[i]

################################ Main ########################################

if len(sys.argv) != 2:
  print "Usage: em.py data"
  sys.exit(1)

data = load_csv(sys.argv[1])
data = cluster.vq.whiten(data)
dump_csv(data, sys.argv[1] + '-3d')

## EM

fem_scores = open("em_scores", "w")
fkmeans_scores = open("kmeans_scores", "w")

for k in range(1, 50):
  print "\n\nk = ", k
  print "\n====="
  alphas = array([1. / k] * k)
  indices = distinct_randint_vals(k, 0, len(data) - 1)
  means = []
  for i in indices:
    means.append(data[i])
  means = array(means)

  start = time.clock()
  kmeans_kernels = deepcopy(means)
  kmeans(data, kmeans_kernels, 5)
  t = time.clock() - start
  print "kmeans:", t, "s"

  variations = [mat('[1., 0. ; 0., 1.]')] * k
  start = time.clock()
  for it in range(0, 10):
    em(data, alphas, means, variations)
  t = time.clock() - start
  print "em:", t, "s"
  
  print "\nem score:", distance_sum(data, means)
  print "kmeans score:", distance_sum(data, kmeans_kernels)

  fem_scores.write(str(k) + " " + str(distance_sum(data, means)) + "\n")
  fkmeans_scores.write(str(k) + " "
                       + str(distance_sum(data, kmeans_kernels)) + "\n")

  fem_scores.flush()
  fkmeans_scores.flush()

  dump_csv(kmeans_kernels, 'kmeans_result' + str(k))
  dump_csv(means, 'em_result' + str(k))

fem_scores.close()
fkmeans_scores.close()

# dump_func(Bind(gaussian, [means[0], variations[0]]),
#           mat('[0., 10. ; 0., 10.]'),
# #           mat([[means[0][0] - 1., means[0][0] + 1.],
# #                [means[0][1] - 1., means[0][1] + 1.]]),
#           [0.5] * 2,
#           "gaussian0")

# # dump_func(Bind(gaussian, [means[1], variations[1]]),
# #           mat([[means[1][0] - 1., means[1][0] + 1.],
# #                [means[1][1] - 1., means[1][1] + 1.]]),
# #           [0.2] * 2,
# #           "gaussian1")

# # dump_func(Bind(gaussian, [means[2], variations[2]]),
# #           mat([[means[2][0] - 1., means[2][0] + 1.],
# #                [means[2][1] - 1., means[2][1] + 1.]]),
# #           [0.2] * 2,
# #           "gaussian2")

# dump_csv(kmeans_result, 'kmeans_result')
# dump_csv(means, 'em_result')
